; Copyright (c) 2004-2010, Laurent Jorda, Stubbe F. Hviid
;   All rights reserves
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; NAME:
;       QUATERNION_TO_EQU
;
; PROJECT:
;       OSIRIS
;
; PURPOSE:
;       This routine computes the equatorial coordinates of the line-of-sight
;       from the components of the quaternion (assuming ESA's definition of
;       the quaternion).
;
; CALLING SEQUENCE:
;       QUATERNION_TO_EQU, q, alf, del, rol, DEBUG=debug, IFAIL=ifail
;
;
; INPUTS:
;       q
;         This variable contains the four components of ESA's quaternion.
;         The three first components correspond to the imaginary (I,J,K) part
;         of the quaternion and the last component corresponds to the real
;         part of the quaternion.
;
; OUTPUTS:
;       alf
;         Right ascension corresponding to the quaternion (rad).
;
;       del
;         Declination corresponding to the quaternion (rad).
;
;       rol
;         Roll angle corresponding to the quaternion (rad). The roll angle is
;         here the angle between the CCD top direction (up) and the direction
;         of the equatorial North (the angle is positive counterclockwise in
;         the image frame). In the image frame for level 1 and level 2 data,
;         West is right when North is up.
;
; KEYWORD PARAMETERS:
;       debug
;         Set to select debug mode. Optional.
;
;       ifail
;         Return the integer error code: 0=OK, >0=error occured
;
; DEPENDENCIES:
;       NONE.
;
; REMARKS:
;       NONE.
;
; MODIFICATION HISTORY:
;       2007/02/09: written by L. Jorda (LAM)
;------------------------------------------------------------------------------
PRO QUATERNION_TO_EQU, q, alf, del, rol, DEBUG=debug, IFAIL=ifail

;
; initialization
;
ct = 180.0d0/!dpi
d  = keyword_set(debug)

if (d) then print,'M: QUATERNION_TO_EQU_01 - BEGIN'

;
; default values
;
alf = 0.0d0
del = 0.0d0
rol = 0.0d0

;
; check input parameters
;
if (n_elements(q) ne 4) then begin

    if (d) then print,'E:Incorrect quaternion!'

    return

endif

;
; display components of quaternion
;
if (d) then begin

    print,'M:Quaterion component #1 = ',q(0)
    print,'M:Quaterion component #2 = ',q(1)
    print,'M:Quaterion component #3 = ',q(2)
    print,'M:Quaterion component #4 = ',q(3)

endif

;
; compute declination 'del' and its trigonometric lines
;
sind = -q(0)^2-q(1)^2+q(2)^2+q(3)^2

if (abs(sind) le 1.0d0) then begin

    del  = asin(sind)
    cosd = sqrt(1.0d0-sind^2)

endif else begin

    del  = !dpi/2.0d0*sind/abs(sind)
    sind = sind/abs(sind)
    cosd = 0.0d0

endelse

if (d) then print,'M:Declination     = ',del*ct,' deg.'

;
; compute right ascension 'alf'
;
cosa = 2.0d0*(q(0)*q(2)-q(1)*q(3))
sina = 2.0d0*(q(1)*q(2)+q(0)*q(3))
alf  = atan(sina,cosa)

if (alf lt 0.0d0) then alf = alf+2.0d0*!dpi

if (d) then print,'M:Right ascension = ',alf*ct,' deg.'

;
; compute roll angle 'rol'
;
cosr = -2.0d0*(q(1)*q(3)+q(0)*q(2))
sinr = -2.0d0*(q(1)*q(2)-q(0)*q(3))
rol  = !dpi-atan(sinr,cosr)

if (rol lt 0.0d0) then rol = rol+2.0d0*!dpi

if (d) then print,'M:Roll angle      = ',alf*ct,' deg.'

;
; normal end of routine
;
if (d) then print,'M: QUATERNION_TO_EQU_01 - END'

RETURN
END


