; Copyright (c) 2004, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   FUNCTION osiris_badpixel_removal, header, image, quality_map, sigma_map
;
; PURPOSE:
;   Bad pixel removal from an Osiris image
;
; CATEGORY:
;   Osiris CCD Calibration
;
; CALLING SEQUENCE:
;  Result = osiris_badpixel_removal(header, image, quality_map, sigma_map)
;
; INPUTS:
;   header:   		The PDS image header of the image to be corrected
;   image:      	The image data of the image to be corrected
;   quality_map:	Existing map of quality estimate for each pixel to be modified
;   sigma_map:		Existing map of estimaped relative error for each pixel (in %) to be modified
;
; KEYWORD PARAMETERS:
;   PARAMETER_FILELIST:	all external datafile names neede for this correction step
;				        will be added to this keywor parameter in a string array
;   DISABLE_FILTER:		Is set the dunction will not modify the image data but will mark the relevant pixels in the quality_map
;
; OUTPUTS:
;   header:   		The PDS image header of the image after correction
;   image:      	The image data of the image after correction
;   quality_map:	output map of quality estimate for each pixel
;   sigma_map:		output map of estimaped relative error for each pixel (in %)
;
; RETURNS:
;   0 on failure   1 on success
;
; RESTRICTIONS:
;   function requires that the paths defined in get_osiris_environment.pro has been set to correctly
;   point to the root directory of the external calibration files
;
; EXAMPLE:
;   p_read, pickfile(), h, d             ; read in raw PDS image
;   osiris_badpixel_removal, h, d
;
; MODIFICATION HISTORY:
;   Written by:  Stubbe F. Hviid   28/6-2005
;                Sfh 02102006 - Switched to using osiris_model for detecting the camera and hardware model
;-

PRO remove_hot_column, image, xoff, column, binning_factor

	s = size(image, /DIM)

	if column GE xoff AND column LT (xoff + s[1]*binning_factor) then begin

		devc = (column - xoff) / binning_factor

		if devc eq 0 then begin
			diff = median(image[0, *]) - median(image[1, *])
			image[0, *] = image[0, *] - diff
		endif else if devc eq (s[0]-1) then begin
			diff = median(image[s[0]-1]) - median(image[s[0]-2])
			image[s[0]-1, *] = image[s[0]-1, *] - diff
		endif else if devc lt s[0] then begin
			diff = ((median(image[devc, *]) - median(image[devc-1, *])) + (median(image[devc, *]) - median(image[devc+1, *])))/2.0

			image[devc, *] = image[devc, *] - diff
		endif
	endif
END

FUNCTION osiris_badpixel_removal, header, image, quality_map, sigma_map, DISABLE_FILTER=DISABLE_FILTER, PARAMETER_FILELIST=PARAMETER_FILELIST, _EXTRA=_EXTRA
VERSION='V1.0'

    if keyword_set(VERBOSE) then print, 'osiris_badpixel_removal running'

    ; Verify the the correction has not already been performed
    flag = p_value(header, 'LEVEL2_PIPELINE.BAD_PIXEL_REPLACEMENT_FLAG', /HISTORY)

    if flag eq 'TRUE' then begin
       if keyword_set(VERBOSE) then print, 'osiris_badpixel_removal skipped - Already performed'
       return, 1
    endif

    ; do the calibration

	; read some header information
	osiris_model, header, camera, model

	bin = p_value(header, 'SR_ACQUIRE_OPTIONS.HARDWARE_BINNING_ID')
	swbin = (p_value(header, 'SR_COMPRESSION.PIXEL_AVERAGING_WIDTH'))[0]
	xoff = p_value(header, 'IMAGE.FIRST_LINE_SAMPLE')-1
	yoff = p_value(header, 'IMAGE.FIRST_LINE')-1
	s = size(image, /DIM)

	; first find database file
	param_file = p_resolve_database_file('BADPIXELS', camera, model, 'BPIX', '', 'txt')
	if param_file eq '' then begin
		print, 'Cannot find required parameter file - osiris_badpixel_removal'
		return, 0
	endif

	; add the loaded parameter file to the PARAMETER_FILELIST
	if n_elements(PARAMETER_FILELIST) eq 0 then PARAMETER_FILELIST = [param_file] else PARAMETER_FILELIST = [PARAMETER_FILELIST, param_file]

	; detemine active binning factor
	bf = 1
	if bin eq '2x2' then bf = 2
	if bin eq '4x4' then bf = 4
	if bin eq '8x8' then bf = 8

	if swbin ne 1 then bf = bf * swbin

	; get the active quality map values
	qmv = osiris_quality_values();


	if camera eq 'NAC' then begin

		; hot column 994 - left filter
		x = (994 - xoff) / bf
		if x ge 0 AND x lt s[0] then begin
			quality_map[x, *] = quality_map[x, *] OR qmv.warm_pixel
			if NOT keyword_set(DISABLE_FILTER) AND bf eq 1 AND x gt 0 then begin
				image[x, *] -= (mean(image[x, *]) - mean(image[x-1, *]))
			endif
		endif

		; hot column 996 - right filter
		x = (996 - xoff) / bf
		if x ge 0 AND x lt s[0] then begin
			quality_map[x, *] = quality_map[x, *] OR qmv.warm_pixel
			if NOT keyword_set(DISABLE_FILTER) AND bf eq 1 AND x lt s[0]-1 then begin
				image[x, *] -= (mean(image[x, *]) - mean(image[x+1, *]))
			endif
		endif

		; bad column 995 - avg filter
		x = (995 - xoff) / bf
		if x ge 0 AND x lt s[0] then begin
			quality_map[x, *] = quality_map[x, *] OR qmv.bad_pixel
			if NOT keyword_set(DISABLE_FILTER) AND bf eq 1 AND x gt 0 AND x lt s[0]-1 then image[x, *] = (image[x-1, *] + image[x+1, *])/2.0
		endif

		; blob
		x = (924 - xoff) / bf
		y = (978 - yoff) / bf
		if x ge 0 AND x lt s[0] AND y ge 0 AND y lt s[1] then begin
			xf = (x-1) > 0
			xt = (x+1) < (s[0]-1)
			yf = (y-1) > 0
			yt = (y+1) < (s[1]-1)
			for tx = xf, xt do begin
				for ty = yf, yt do begin
					quality_map[tx, ty] = qmv.bad_pixel
				endfor
			endfor
		endif

		if NOT keyword_set(DISABLE_FILTER) then begin
		if x gt 1 AND x lt s[0]-2 then begin
			if y gt 1 AND x lt s[1]-2 then begin
				image[x-1, y-1] = (image[x-2, y-1] + image[x-2, y-2] + image[x-1, y-2]) / 3.0		; -1, -1
				image[x,   y-1] = image[x, y-2]    													; 0, -1
				image[x+1, y-1] = (image[x+2, y-1] + image[x+2, y-2] + image[x+1, y-2]) / 3.0		; +1, -1
				image[x-1, y]   = image[x-2, y]														; -1, 0
				image[x+1, y]   = image[x+2, y]														; +1, 0

				image[x-1, y+1] = (image[x-2, y+1] + image[x-2, y+2] + image[x-1, y+2]) / 3.0		; -1, +1
				image[x,   y+1] = image[x, y+2]    													; 0, +1
				image[x+1, y+1] = (image[x+2, y+1] + image[x+2, y+2] + image[x+1, y+2]) / 3.0		; +1, +1
				image[x, y] = (image[x-1, y] + image[x+1, y] + image[x, y-1] + image[x, y+1]) / 4.0	; 0,0
			endif
		endif

		endif
			; center - bad 924, 978

			; surrounding 8 pixels warm

	endif

	if camera eq 'WAC' then begin

		; dim column segment x: 119 y: [1664, 2047]

		x = (119 - xoff) / bf
		y = ([1664, 2047] - yoff) / bf

		if x gt 0 AND x le s[0] then begin
			yf = y[0] > 0 < s[1]
			yt = y[1] > 0 < s[1]

			if yf ge 0 AND yf lt s[1] AND yt ge 0 AND yt lt s[1] then begin
				quality_map[x, yf:yt] = quality_map[x, yf:yt] OR (qmv.dim_pixel OR qmv.bad_pixel)
				if NOT keyword_set(DISABLE_FILTER) then begin
					if x gt 0 AND x lt s[0]-1 then begin
						image[x, yf:yt] = (image[x-1, yf:yt] + image[x+1, yf:yt])/2.0
					endif
				endif
			endif
		endif



	endif


	p_approc, header, header, 'osiris_badpixel_removal', VERSION, 'Using embedded table'


    ; Change the processing flag
    if NOT keyword_set(DISABLE_FILTER) then p_setvalue, header, 'LEVEL2_PIPELINE.BAD_PIXEL_REPLACEMENT_FLAG', 'TRUE', /LABEL, /NO_QUOTES, /HISTORY
    p_setvalue, header, 'LEVEL2_PIPELINE.BAD_PIXEL_REPLACEMENT_VERSION', VERSION, /GROUP, /HISTORY


    ; finish
    if keyword_set(VERBOSE) then print, 'osiris_badpixel_removal - Done'

    return, 1
END