; Copyright (c) 2004-2010, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   FUNCTION osiris_calibrate_adc_offset, header, image, quality_map, sigma_map
;
; PURPOSE:
;   Correct TANDEM ADC offset shift between the two ADC channels
;
; CATEGORY:
;   Osiris CCD Calibration
;
; CALLING SEQUENCE:
;  Result = osiris_calibrate_adc_offset(header, image, quality_map, sigma_map)
;
; INPUTS:
;   header:   		The PDS image header of the image to be corrected
;   image:      	The image data of the image to be corrected
;   quality_map:	Existing map of quality estimate for each pixel to be modified
;   sigma_map:		Existing map of estimaped relative error for each pixel (in %) to be modified
;
; KEYWORD PARAMETERS:
;   PARAMETER_FILELIST: all external datafile names neede for this correction step
;                  will be added to this keywor parameter in a string array
;
; OUTPUTS:
;   header:   		The PDS image header of the image after correction
;   image:      	The image data of the image after correction
;   quality_map:	output map of quality estimate for each pixel
;   sigma_map:		output map of estimaped relative error for each pixel (in %)
;
; RETURNS:
;   0 on failure   1 on success
;
; RESTRICTIONS:
;   function requires that the paths defined in get_osiris_environment.pro has been set to correctly
;   point to the root directory of the external calibration files
;
;   The function depends on an external data file <camera>_<unit>_ADC-OFFSET_V<version>_<timestamp>.dat
;   to exist in the ADC folder of the database tree. This file must have a tag_file compatible structure
;   with minimum the tag 'DIF' existing as a 3-element integer vector
;
; EXAMPLE:
;   p_read, pickfile(), h, d             ; read in raw PDS image
;   osiris_calibrate_adc, h, d
;
; MODIFICATION HISTORY:
;   Written by:  Stubbe F. Hviid   12/8-2004 based on p_adc_hi.pro by R. Kramm dated 29-OCT-2003
;                Slightly changed condition to treat the image at all, MK, 16 09 2005
;                Sfh 02102006 - Switched to using osiris_model for detecting the camera and hardware model
;-

FUNCTION osiris_calibrate_adc_offset, header, image, quality_map, sigma_map, PARAMETER_FILELIST=PARAMETER_FILELIST, _EXTRA=_EXTRA
VERSION='V1.1'

    if keyword_set(VERBOSE) then print, 'osiris_calibrate_adc_offset running'

    ; Verify the the correction has not already been performed
    flag = p_value(header, 'LEVEL2_PIPELINE.ROSETTA:ADC_OFFSET_CORRECTION_FLAG', /HISTORY)

    if flag eq 'TRUE' then begin
       if keyword_set(VERBOSE) then print, 'osiris_calibrate_adc_offset skipped - Already performed'
       return, 1
    endif

    ; do the calibration

    ; extract some information from the image header
    osiris_model, header, camera, model

    amp = p_value(header, 'SR_ACQUIRE_OPTIONS.AMPLIFIER_ID')
    adc = p_value(header, 'SR_ACQUIRE_OPTIONS.ADC_ID')


    ; Check if we need to do anything. This is always the case for ADC high and for ADC tandem if
    ; pixel values ge 2^14 are reached.

    if (adc eq 'HIGH') OR ((adc eq 'TANDEM') AND (max(image) ge 2L^14)) then begin

       ; load calibration parameters from database
       param_file = p_resolve_database_file('ADC', camera, model, 'ADC', '', '*')

       if param_file eq '' then begin
         printf, -2, 'ERROR: Cannot find required parameter file - osiris_calibrate_adc_offset'
         return, 0
       endif
       params = fw_load_tagfile(param_file)

       ; add the loaded parameter file to the PARAMETER_FILELIST
       if n_elements(PARAMETER_FILELIST) eq 0 then PARAMETER_FILELIST = [param_file] else PARAMETER_FILELIST = [PARAMETER_FILELIST, param_file]

       ; extract the dif parameter vector
       dif = [params.adc_offset_a, params.adc_offset_b, params.adc_delta_ab]

       ; determine the limit
       if adc eq 'TANDEM' then lim = 2L^14-1L else lim = -1L

       ; handle AB amplifier chain
       if amp eq 'A' then begin
         ; define processing message
         proc_message = 'Amplifier A ADC offset correction needed (Dif=' + strtrim(dif[0], 2) + ')'
         ; find pixels that requires correction
         s = where(image gt lim)
         ; Subtract the offset
         image[s] = image[s] - dif[0]
       endif else if amp eq 'B' then begin
         ; define processing message
         proc_message = 'Amplifier B ADC offset correction needed (Dif=' + strtrim(dif[1], 2) + ')'
         ; find pixels that requires correction
         s = where(image gt lim)
         ; Subtract the offset
         image[s] = image[s] - dif[1]
       endif else if amp eq 'BOTH' then begin

         ; correct dif for the amplifier BOTH mode
         dif = long(dif[0:1] + dif[2])

         ; define processing message
         proc_message = 'Amplifier BOTH ADC offset correction needed (Dif=A:' + strtrim(dif[0], 2) + ', B:' + strtrim(dif[1], 2) + ')'

         ; find image size
         dim = size(image, /DIM)

         ; extract sub image for each amplifier
         image_a = image[0:dim[0]/2 - 1, *]
         image_b = image[dim[0]/2: dim[0]-1, *]

         ; handle amp A segment
         sa = where(image_a gt lim)
         if sa[0] ge 0 then image_a[sa] = image_a[sa] - dif[0]

         ; handle amp B segment
         sb = where(image_b gt lim)
         if sb[0] ge 0 then image_b[sb] = image_b[sb] - dif[1]

         ; recombine image
         image[0:dim[0]/2 - 1, *] = image_a
         image[dim[0]/2: dim[0]-1, *] = image_b

       endif else begin
         printf, -2, 'ERROR: Illegal amplifier ID: ' + amp + ' was found - osiris_calibrate_adc_offset'
         return, 0
       endelse
    endif else proc_message = 'No ADC offset correction needed'


    ; Change the processing flag
    p_setvalue, header, 'LEVEL2_PIPELINE.ADC_OFFSET_CORRECTION_FLAG', 'TRUE', /LABEL, /NO_QUOTES, /HISTORY
    p_setvalue, header, 'LEVEL2_PIPELINE.ADC_OFFSET_CORRECTION_VERSION', VERSION, /GROUP, /HISTORY


    ; finish
    if keyword_set(VERBOSE) then print, 'osiris_calibrate_adc_offset - Done'

    return, 1
END