; Copyright (c) 2004-2010, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   FUNCTION osiris_level1_to_level2, header_in, image_in, header_out, image_out, quality_map, sigma_map
;
; PURPOSE:
;   Main interface function to the CODMAC level 1 to level 2 image converter
;   The function will convert from raw DN based imaged to radiometrically calibrated level 2 images
;   in units of W/m^2/sr/nm
;
; CATEGORY:
;   Osiris CCD Calibration
;
; CALLING SEQUENCE:
;  Result = osiris_level1_to_level2(header_in, image_in, header_out, image_out, quality_map, sigma_map)
;
; INPUTS:
;   header_in:   input level 1 PDS header structure
;   image_in:  input level 1 image data
;
; KEYWORD PARAMETERS:
;   SKIP_ABS: If set, absolute calibration is skipped.
;   SKIP_EXPTIME: If set, division by exposure time is skipped. Should normally be used only when
;                 SKIP_ABS is set, too.
;   SKIP_FLATFIELD: If set, division by flatfield is not performed.
;   USE_NOISE_CAL:  If set the coherent noise calibration will be performed if not set the calibration step is skipped
;   USE_DARK_CAL:   If set the primitive dark current calibration will be performed if not set the calibration step is skipped
;   CODMAC:			If set the use CODMAC data level else use OSIRIS data levels
;
; OUTPUTS:
;   header_out:     	output PDS image header
;   image_out:      	output calibrated image data
;   quality_map_out:	output map of quality estimate for each pixel
;   sigma_map_out:		output map of estimaped relative error for each pixel (in %)
;
; RETURNS:
;   0 on failure   1 on success
;
; RESTRICTIONS:
;   function requires that the paths defined in get_osiris_environment.pro has been set to correctly
;   point to the root directory of the external calibration files
;
; EXAMPLE:
;   p_read, pickfile(), h_in, d_in                 ; read in raw PDS image
;   osiris_level1_to_level2, h_in, d_in, h_out, d_out ; calibrate the image
;   p_write, pickfile(), h_out, d_out             ; save the image to a new PDS file
;
; MODIFICATION HISTORY:
;   Written by:  Stubbe F. Hviid   12/8-2004
;                Added Keyword SKIP_EXPTIME, MK, 20 Oct. 2005
;                Added Keyword SKIP_FLATFIELD, MK, 02 Nov. 2005
;                Propagate Keyword SKIP_FLATFIELD to osiris_can_calibrate, MK, 25 Jan. 2009
;-

FUNCTION osiris_level1_to_level2, header_in, image_in, header_out, image_out, quality_map_out, sigma_map_out, _EXTRA=_EXTRA, $
                                  SKIP_ABS=SKIP_ABS,SKIP_EXPTIME=SKIP_EXPTIME,SKIP_FLATFIELD=SKIP_FLATFIELD, $
                                  USE_NOISE_CAL=USE_NOISE_CAL, USE_DARK_CAL=USE_DARK_CAL, $
                                  CODMAC=CODMAC
VERSION = '1.5'

    ; initiate status variable
    ok = 1

    ; start by setting output equal to the input
    header_out = header_in
    image_out = double(image_in)   ; use double presition floating point

	; initiate the quality and sigma maps
	s = size(image_out, /DIM)
	quality_map_out = bytarr(s[0], s[1])
	sigma_map_out = fltarr(s[0], s[1])


	; Make quick check to see if it at all makes sense to continue with the calibration
	; (This speed things up quite a log when doing batch processing)
	if ok ne 0 then begin
		ok = osiris_can_calibrate(header_out, SKIP_FLATFIELD=SKIP_FLATFIELD) ;keyword skip_flatfield added by MK

		if ok eq 0 then printf, -2, 'Skipping ' + header_in.filename + ' from calibration - failed osiris_can_calibrate'
	endif

	; generate primary quality map
	if ok ne 0 then ok = osiris_mk_quality_map(header_out, image_out, quality_map_out)

    ; Deconvolve TANDEM ADC offset
    if ok ne 0 then ok = osiris_calibrate_adc_offset(header_out, image_out, quality_map_out, sigma_map_out, PARAMETER_FILELIST=PARAMETER_FILELIST, _EXTRA=_EXTRA)

    ; Remove coherent noise (Commented out as per Kramm recommendation)
    if keyword_set(USE_NOISE_CAL) then begin
    	if ok ne 0 then ok = osiris_calibrate_noise(header_out, image_out, quality_map_out, sigma_map_out)
    endif

    ; seperate out the various frame elements
    if ok ne 0 then ok = osiris_seperate_frames(header_out, image_out, quality_map_out, sigma_map_out, prepix, postpix, ovrclklines)

	; check if any ccd image dtaa is present
	if image_out[0] eq -1 then begin
		printf, -2, 'ERROR - No CCD data in image, only pre and post pixels -cannot calibrate'
		return, 0
	endif

    ; remove the bias offset
    if ok ne 0 then ok = osiris_subtract_bias(header_out, image_out, quality_map_out, sigma_map_out, prepix, postpix, PARAMETER_FILELIST=PARAMETER_FILELIST, _EXTRA=_EXTRA)

	; build primary error map
	if ok ne 0 then ok = osiris_calc_poisson(header_out, image_out, quality_map_out, sigma_map_out)

    ; Convert to DN/s
    if NOT keyword_set(SKIP_EXPTIME) then $
       if ok ne 0 then ok = osiris_exposuretime(header_out, image_out, quality_map_out, sigma_map_out, PARAMETER_FILELIST=PARAMETER_FILELIST, _EXTRA=_EXTRA)

    ; remove darkcurrent
    if keyword_set(USE_DARK_CAL) then begin
    	if ok ne 0 then ok = osiris_darkcurrent_removal(header_out, image_out, quality_map_out, sigma_map_out, PARAMETER_FILELIST=PARAMETER_FILELIST, _EXTRA=_EXTRA)
    endif

    ; Remove bad pixels and bad columns
    if ok ne 0 then ok = osiris_badpixel_removal(header_out, image_out, quality_map_out, sigma_map_out, PARAMETER_FILELIST=PARAMETER_FILELIST, _EXTRA=_EXTRA)

    ; Flatfield correct the image
    if NOT keyword_set(SKIP_FLATFIELD) then $
         if ok ne 0 then ok = osiris_flatfield(header_out, image_out, quality_map_out, sigma_map_out, PARAMETER_FILELIST=PARAMETER_FILELIST, _EXTRA=_EXTRA)

    ; Finally convert to radiencel units
    if NOT keyword_set(SKIP_ABS) then $
       if ok ne 0 then ok = osiris_abscal(header_out, image_out, quality_map_out, sigma_map_out, PARAMETER_FILELIST=PARAMETER_FILELIST, _EXTRA=_EXTRA)

    ; Save a list of used parameter/calibration files to the header
    if n_elements(PARAMETER_FILELIST) eq 0 then begin
       p_setvalue, header_out, 'LEVEL2_PIPELINE.CALIBRATION_DATAFILES', 'NULL', /LABEL, /NO_QUOTES, /GROUP, /HISTORY
    endif else begin
       ; first strip paths from the parameter files
       param_filelist = strarr(n_elements(PARAMETER_FILELIST))
       for i=0, n_elements(PARAMETER_FILELIST)-1 do begin
         fp = p_parse_file_path(PARAMETER_FILELIST[i])
         param_filelist[i] = fp.filename
       endfor
       p_setvalue, header_out, 'LEVEL2_PIPELINE.CALIBRATION_DATAFILES', param_filelist, /GROUP, /HISTORY
    endelse

    ; update header
    p_approc,header_out,header_out,'osiris_level1_to_level2','V'+VERSION
    p_setvalue, header_out, 'LEVEL2_PIPELINE.PIPELINE_MASTER_VERSION', VERSION, /GROUP, /HISTORY

    ; delete object that are no longer relevant in the RDR header
    if p_is_group(header_out, 'BLADE1_PULSE_ARRAY') then begin
       p_delete_group, header_out, 'BLADE1_PULSE_ARRAY'
       p_delete_tag, header_out, '^BLADE1_PULSE_ARRAY'
    endif
    if p_is_group(header_out, 'BLADE2_PULSE_ARRAY') then begin
       p_delete_group, header_out, 'BLADE2_PULSE_ARRAY'
       p_delete_tag, header_out, '^BLADE2_PULSE_ARRAY'
    endif
    if p_is_group(header_out, 'IMAGE_HEADER') then begin
       p_delete_group, header_out, 'IMAGE_HEADER'
       p_delete_tag, header_out, '^IMAGE_HEADER'
    endif
    if p_is_group(header_out, 'PA_IMAGE') then begin
       p_delete_group, header_out, 'PA_IMAGE'
       p_delete_tag, header_out, '^PA_IMAGE'
    endif
    if p_is_group(header_out, 'PB_IMAGE') then begin
       p_delete_group, header_out, 'PB_IMAGE'
       p_delete_tag, header_out, '^PB_IMAGE'
    endif
    if p_is_group(header_out, 'OL_IMAGE') then begin
       p_delete_group, header_out, 'OL_IMAGE'
       p_delete_tag, header_out, '^OL_IMAGE'
    endif

    ; change the product type
    if keyword_set(SKIP_ABS) then begin
       if keyword_set(SKIP_EXPTIME) then p_setvalue, header_out, 'IMAGE.UNIT', 'DU' else $
                                         p_setvalue, header_out, 'IMAGE.UNIT', 'DU/s'
    endif else begin
       p_setvalue, header_out, 'IMAGE.UNIT', 'W/[M^2 SR NM]'
    endelse

    p_setvalue, header_out, 'PRODUCT_TYPE', 'RDR'
    if keyword_set(CODMAC) then begin
    	p_setvalue, header_out, 'PROCESSING_LEVEL_ID', '3'	; Setting CODMAC data level 3
    	dsi = p_value(header_out, 'DATA_SET_ID')
    	dsi = strmid(dsi, 0, 12) + '3' + strmid(dsi, 13)
    	p_setvalue, header_out, 'DATA_SET_ID', dsi

		dsn = p_value(header_out, 'DATA_SET_NAME')
		p = strpos(dsn, 'OSI')
		if p gt 0 then dsn = strmid(dsn, 0, p+7) + '3 RDR data'
		p_setvalue, header_out, 'DATA_SET_NAME', dsn
    endif else begin
    	p_setvalue, header_out, 'PROCESSING_LEVEL_ID', 2	; Setting OSIRIS data level 2
    	dsi = p_value(header_out, 'DATA_SET_ID')
    	dsi = strmid(dsi, 0, 5) + 'OSIRIS-2' + strmid(dsi, 13)
    	p_setvalue, header_out, 'DATA_SET_ID', dsi

		dsn = p_value(header_out, 'DATA_SET_NAME')
		p = strpos(dsn, 'OSI')
		if p gt 0 then dsn = strmid(dsn, 0, p) + ' OSIRIS 2 RDR data'
		p_setvalue, header_out, 'DATA_SET_NAME', dsn
	endelse
    p_setvalue, header_out, 'PROCESSING_LEVEL_DESC', 'Radiometrically Calibrated Data'
    ;p_setvalue, header_out, 'PRODUCER_FULL_NAME', strtrim(getenv('USERNAME'), 2)
    p_setvalue, header_out, 'PRODUCT_CREATION_TIME', p_timenow(), /NO_QUOTES, /LABEL


	; filter out any infinite or NaN values
	s = where(NOT finite(image_out), count)
	if count then begin
		printf, -2, 'WARNING: Nan or INF values found in output image!'
		image_out[s] = 0.0	; set any Nan or INF to 0.0 value

		qmv = osiris_quality_values()
		quality_map_out[s] = quality_map_out[s] OR qmv.bad_pixel	; and mark the pixels bad
	endif

	; update the DATA_QUALITY_ID field
	dqid_h = strtrim( string(ulong(mean(quality_map_out)), FORMAT='(%"%03d")' ), 2)
	dqid_l = strtrim( string(ulong(max(quality_map_out)), FORMAT='(%"%03d")' ), 2)

	p_setvalue, header_out, 'DATA_QUALITY_ID', strtrim((dqid_h + dqid_l), 2)
	p_setvalue, header_out, 'DATA_QUALITY_DESC', 'the top 3 digits are the mean of the quality mapand the bottom three are the max'

    ; finish
    return, ok
END