; Copyright (c) 2004-2010, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   FUNCTION osiris_separate_frames, header, image, pre_pix, post_pix, ovrclklines
;
; PURPOSE:
;   Seperate the various frame elements in a osiris image (prepixels/post pixels, image frame, overclocking lines.
;
; CATEGORY:
;   Osiris CCD Calibration
;
; CALLING SEQUENCE:
;  Result = osiris_separate_frames(header, image, quality_map, sigma_map, pre_pix, post_pix, ovrclklines)
;
; INPUTS:
;   header:   		The PDS image header of the image to be corrected
;   image:    		The image data of the image to be corrected
;   quality_map:	Existing map of quality estimate for each pixel to be modified
;   sigma_map:		Existing map of estimaped relative error for each pixel (in %) to be modified
;
; KEYWORD PARAMETERS:
;   NONE
;
; OUTPUTS:
;   header:   		The PDS image header of the image after correction
;   image:    		Will contain the image frame (CCD surface) after seperation
;   quality_map:	output map of quality estimate for each pixel
;   sigma_map:		output map of estimaped relative error for each pixel (in %)
;   prepix:   		Will contain the image pre pixels
;   postpix:  		Will contain the image post pixels
;   ovrclklines: 	Will contain the image overclocking lines
;
;   Note! that in case of amplifier BOTH images the prepix will contain the prepixels of the A amplifier
;   and postpix will contains the prepixels of the B amplifier
;
;  If a certain frame segment is not present in the original image data the the corresponding output field
;  will be -1
;
; RETURNS:
;   0 on failure   1 on success
;
; RESTRICTIONS:
;   function requires that the paths defined in get_osiris_environment.pro has been set to correctly
;   point to the root directory of the external calibration files.
;
;
; EXAMPLE:
;   p_read, pickfile(), h, d             ; read in raw PDS image
;   print, osiris_separate_frames(h, d, pre_pix, post_pix, ovrclklines)
;
; MODIFICATION HISTORY:
;   Written by:  Stubbe F. Hviid   27/6-2005
;   Modified by MK to handle prepixels correctly for amplifier B, MK, 2008
;-

FUNCTION osiris_seperate_frames, header, image, quality_map, sigma_map, prepix, postpix, ovrclklines
VERSION = 'V1.0'

	if keyword_set(VERBOSE) then print, 'osiris_separate_frames running'  ;corrected routine name, MK

	; extract the image size
	dim = size(image, /DIM)

	; extract some information from the image header
	amp    = p_value(header, 'SR_ACQUIRE_OPTIONS.AMPLIFIER_ID')
	bin    = p_value(header, 'SR_ACQUIRE_OPTIONS.HARDWARE_BINNING_ID')
	sw_bin = p_value(header, 'SR_COMPRESSION.PIXEL_AVERAGING_WIDTH')

	; extract frame offset
	xf = p_value(header, 'IMAGE.FIRST_LINE_SAMPLE')-1
	yf = p_value(header, 'IMAGE.FIRST_LINE')-1
	w  = p_value(header, 'IMAGE.LINE_SAMPLES')
	h  = p_value(header, 'IMAGE.LINES')


	; determine effective binning factor
	bf = 1;

	; first handle the hardware binning
	if bin eq '2x2' then bf = 2;
	if bin eq '4x4' then bf = 4;
	if bin eq '8x8' then bf = 8;

	; then the software binning
	bf = bf * sw_bin[0]


	xf = xf/bf
	yf = yf/bf

	ccd_size = 2048 / bf

	; handle overclocking lines first
	if yf + h gt ccd_size then begin
		ol = (yf + h) - ccd_size
		ovrclklines = image[h-ol:h-1]
		image = image[*, 0:h-ol-1]
		h = h - ol
	endif else ovrclklines = [-1]

	; then handle pre pixels.
	; MK: Changed prepixel handling to consider amplifier for prepixels. The issue is that, for Amplifier B with prepixels,
	; FIRST_LINE_SAMPLE is still -48, but the prepixels are at the end of the image ([2048:2095,*] for a full frame)
	if xf lt 0 then begin
		if amp eq 'B' then pp1 = image[w+xf:w-1, *] else pp1 = image[0:-xf-1, *]
    endif else pp1 = [-1]

	if xf + w gt ccd_size then begin
		if xf gt ccd_size then pp2 = image else	pp2 = image[ccd_size-xf:w-1, *]
	endif else pp2 = [-1]

	if pp1[0] ne -1 OR pp2[0] ne -1 then begin
		if xf gt ccd_size then ccd = [-1] else begin
			l = 0
			r = w-1
			if pp1[0] ne -1 then begin
				s = size(pp1, /dim)
				if (amp eq 'B') then r = r-s[0] else l = l + s[0]   ; This was adjusted for Amp B
			endif

			if pp2[0] ne -1 then begin
				s = size(pp2, /dim)
				r = r - s[0]
			endif

			ccd = image[l:r, *]
		endelse
	endif else ccd = image

	; seperate the frames
	if amp eq 'A' then begin
		prepix = pp1
		postpix = pp2
		image = ccd
	endif else if amp eq 'B' then begin
		prepix = pp2
		postpix = pp1
		image = ccd
	endif else begin
		prepix = pp1
		postpix = pp2
		image = ccd
	endelse


	; correct header
	if n_elements(ccd) ne 0 then begin
		if xf lt 0 then	p_setvalue, header, 'IMAGE.FIRST_LINE_SAMPLE', 1
	endif

	; Add processing history entry
	proc_message = 'Found: '
	if image[0] ne -1 then proc_message = proc_message + 'Image Data'
	if prepix[0] ne -1 then proc_message = proc_message + ', Prepixels'
	if postpix[0] ne -1 then proc_message = proc_message + ', Postpixels'
	if ovrclklines[0] ne -1 then proc_message = proc_message + ', Overclocking Lines'

	p_setvalue, header, 'LEVEL2_PIPELINE.FRAME_SEPERATOR_VERSION', VERSION, /GROUP, /HISTORY

	; success at this point in the flow
	return, 1
END