; Copyright (c) 2004-2010, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   FUNCTION osiris_shutter, header
;
; PURPOSE:
;   Generate image matrix with effective exposuretime (excluding correction for binning)
;   for each pixel in the input image
;
; CATEGORY:
;   Osiris CCD Calibration
;
; CALLING SEQUENCE:
;  Result = osiris_shutter(header)
;
; INPUTS:
;   header:   		The PDS image header of the image to be processed
;
; KEYWORD PARAMETERS:
;   PARAMETER_FILELIST:	all external datafile names neede for this correction step
;				        will be added to this keywor parameter in a string array
;   VERSION			    return the software version of the osiris_shutter function
; RETURNS:
;   -1 on error
;   W x H image frame with the calculated exposure time for each pixel
;
; RESTRICTIONS:
;   function requires that the paths defined in get_osiris_environment.pro has been set to correctly
;   point to the root directory of the external calibration files
;
;   The function depends on an external data file <camera>_<unit>_SHM_B#_<version>_<timestamp>.img
;   to exist in the Shutter folder of the database tree.
;
; EXAMPLE:
;   p_read, pickfile(), h, d             ; read in raw PDS image
;   exp_time = osiris_shutter(h)
;   dn_per_sec = float(d) / exp_time
;
; MODIFICATION HISTORY:
;   Written by:  Stubbe F. Hviid   15/6-2009
;                2010-07-28 SFH - Fixed errors with the pulse index calculation
;-

FUNCTION osiris_shutter, header, PARAMETER_FILELIST=PARAMETER_FILELIST, VERBOSE=VERBOSE, VERSION=VERSION, _EXTRA=_EXTRA
VERSION='V2.0'

    if keyword_set(VERBOSE) then print, 'osiris_shutter running'

    ; do the calibration

	; extract some information from the image header
	osiris_model, header, camera, model

	b1_file = p_resolve_database_file('SHUTTER', camera, model, 'SHM-B1', '', '*')
	b2_file = p_resolve_database_file('SHUTTER', camera, model, 'SHM-B2', '', '*')

	if b1_file eq '' OR b2_file eq '' then begin
		printf, -2, 'ERROR:  Cannot find required data files for shutter calibration'
		return, -1
	endif

	tb1 = fw_load_tagfile(b1_file)
	tb2 = fw_load_tagfile(b2_file)

	x = p_value(header, 'IMAGE.FIRST_LINE_SAMPLE')-1
	y = p_value(header, 'IMAGE.FIRST_LINE')-1
	w = p_value(header, 'IMAGE.LINE_SAMPLES')
	h = p_value(header, 'IMAGE.LINES')

	zero_pulse = p_value(header, 'SR_SHUTTER_CONFIG.ZEROPULSE_FLAG')

	hw_bin = p_value(header, 'SR_ACQUIRE_OPTIONS.HARDWARE_BINNING_ID')
	sw_bin = p_value(header, 'SR_COMPRESSION.PIXEL_AVERAGING_WIDTH')

	duration = p_value(header, 'SR_ACQUIRE_OPTIONS.EXPOSURE_DURATION')

	shm_pulse_duration = 0.47619047619e-6 ; s


	; quick sanity check
	if min(sw_bin) ne max(sw_bin) then begin
	print, 'ERROR - Inconsistent software windowing size'
		return, -1
	endif else sw_bin = sw_bin[0]


	; initiate the output matrix
	exp_time = fltarr(2048, 2048)
	exp_time[*,*] = duration

	; read shutter pulse data
	shm = p_read_shm_object(header, OK=ok)


	; handle case where we have shutter pulses
	handled = 0
	if ok ne 0 then begin
		if n_elements(shm.b1) gt 420 AND n_elements(shm.b2) gt 420 then begin
			handled = 1

			; handle zero pulse
			if zero_pulse eq 'TRUE' then begin
				thZ_b1 = tb1.thZ * !DPI / 180.0
				thZ_b2 = tb2.thZ * !DPI / 180.0
			endif else begin
				thZ_b1 = 0.0
				thZ_b2 = 0.0
			endelse

			ccd_size_mm = 27.648
			pixel_size_mm = 13.5e-3

			th0_b1 = tb1.th0 * !DPI / 180.0
			th0_b2 = tb2.th0 * !DPI / 180.0

			rad_per_stripe_b1 = tb1.deg_per_stripe * !DPI / 180.0
			rad_per_stripe_b2 = tb2.deg_per_stripe * !DPI / 180.0


			; calculate travel distance at pulse 0
			T0_b1 = tb1.L * (sin(th0_b1) + sin(thZ_b1 - th0_b1))
			T0_b2 = tb2.L * (sin(th0_b2) + sin(thZ_b2 - th0_b2))

			; calculate the pulse index of the first pulse on the CCD
			n_first_b1 = long((asin(tb1.pt / tb1.L - sin(th0_b1)) - thZ_b1 + th0_b1) / rad_per_stripe_b1)
			n_first_b2 = long((asin(tb2.pt / tb2.L - sin(th0_b2)) - thZ_b2 + th0_b2) / rad_per_stripe_b2)

			; calculate the pulse index of the last pulse on the CCD
			n_last_b1 = long((asin((tb1.pt+ccd_size_mm) / tb1.L - sin(th0_b1)) - thZ_b1 + th0_b1) / rad_per_stripe_b1) + 1
			n_last_b2 = long((asin((tb2.pt+ccd_size_mm) / tb2.L - sin(th0_b2)) - thZ_b2 + th0_b2) / rad_per_stripe_b2) + 1


			; genberate transfer functon between pixel index and pulse index
			pixel_index_b1 = dindgen(2048)
			pixel_index_b2 = dindgen(2048)

			pulse_index_b1 = (asin((pixel_index_b1 * pixel_size_mm + tb1.pt) / tb1.L - sin(th0_b1)) - thZ_b1 + th0_b1) / rad_per_stripe_b1 - n_first_b1
			pulse_index_b2 = (asin((pixel_index_b2 * pixel_size_mm + tb2.pt) / tb2.L - sin(th0_b2)) - thZ_b2 + th0_b2) / rad_per_stripe_b2 - n_first_b2


			; extract the relevant region of the shutter pulse data
			b1 = double(shm.b1[n_first_b1:n_last_b1]) * shm_pulse_duration
			x1 = dindgen(n_elements(b1))
			b2 = double(shm.b1[n_first_b2:n_last_b2]) * shm_pulse_duration
			x2 = dindgen(n_elements(b2))

			; fir the pulse data with a polynomia fit
			fit_b1 = poly_fit(x1, b1, 9, /double)
			fit_b2 = poly_fit(x2, b2, 9, /double)

			; generate exposuretime correction for each pixel

			t1 = poly(pulse_index_b1, fit_b1)
			t2 = poly(pulse_index_b2, fit_b2)


			correction_time = t1 - t2
			;plot, correction_time, yrange=[0.0, 0.005]
			print, 'correction_time mean = ', mean(correction_time)

			for i=0, 2047 do exp_time[i, *] -= correction_time
		endif
	endif else begin
		if p_value(header, 'SR_SHUTTER_STATUS.BLADE1_FIT_SLOPE') ne 'NULL' then begin
			; handle case where we have no shutter pulses but a motion fir by the IFSW

			print, 'WARNING:  shutter fit data available but implementation missing!!'
			handled = 0
		endif
	endelse

	if handled eq 0 then begin
		; handle case where we have no shutter data at all
		print, 'WARNING: No shutter data found in header - using default correction'
		exp_time -= 0.0027
	endif


	bin_factor = 1
	if hw_bin eq '2x2' then bin_factor=2
	if hw_bin eq '4x4' then bin_factor=4
	if hw_bin eq '8x8' then bin_factor=8

	if sw_bin ne 1 then bin_factor = bin_factor * sw_bin

	if bin_factor gt 1 then begin
		w *= bin_factor
		h *= bin_factor
	endif

	; chop subframe out of the global exp_time matrix
	if x ge 0 AND x lt 2048 AND y ge 0 AND y lt 2048 then begin
		xt = x+w-1
		yt = y+h-1

		if xt ge 0 AND xt lt 2048 AND yt ge 0 AND yt lt 2048 then begin
			out = exp_time[x:xt, y:yt]
		endif else stop
	endif else stop




	if bin_factor gt 1 then begin
		;out = congrid(out, w / bin_factor, h / bin_factor, CUBIC=-0.5)
		out = rebin(out, long(w / bin_factor), long(h / bin_factor))
	endif

    ; finish
    if keyword_set(VERBOSE) then print, 'osiris_shutter - Done'

    return, out
END